from __future__ import annotations
from typing import Any, Sequence
import toml
from pathlib import Path
from argparse import ArgumentParser
from .scheduler import HPCScheduler


def parse_config_path_arg() -> str | None:
    parser = ArgumentParser('Schedule multiple jobs or run a single job.')
    # if config is specified, launch single run, else, schedule multi runs
    parser.add_argument('--config', type=str, default=None)
    return parser.parse_args().config


def run(config_path: str | Path) -> None:
    config = toml.load(config_path)
    file = Path(config['run']['file'])
    if file.suffix == '.py':
        file = file.with_suffix('')
    function = config['run']['function']
    args = config['args']
    exec(f'from {file} import {function}')
    eval(f'{function}(**{args})')


def create_configs(
    *,
    # ___________________________________ SLURM #
    working_dir: str,
    jobs_dir: str,
    partition: str,
    account: str,
    node: int,
    task: int,
    cpu: int,
    gpu: int,
    ram: int,
    constraint: str,
    modules: list[str],
    commands: list[str],
    # _____________________________________ JOB #
    file: str,
    function: str,
    args: dict[Any: Any] = dict(),
    # ________________________________ SCHEDULE #
    name: str,
    schedule: dict[str: Sequence[Any]] = dict(),
) -> None:
    # template config
    template_config = dict()
    template_config['run'] = dict(file=file, function=function)
    template_config['args'] = args
    with open((Path(working_dir) / 'template').with_suffix('.toml'), 'w') as config_file:
        toml.dump(template_config, config_file)
    # schedule config
    schedule_config = dict()
    schedule_config['slurm'] = dict(name=name, working_dir=working_dir, jobs_dir=jobs_dir,
                                    partition=partition, account=account,
                                    node=node, task=task, cpu=cpu, gpu=gpu, ram=ram,
                                    constraint=constraint, modules=modules, commands=commands)
    schedule_config['parameters'] = dict()
    for key, values in schedule.items():
        schedule_config['parameters'][key] = values
    with open((Path(working_dir) / 'schedule').with_suffix('.toml'), 'w') as config_file:
        toml.dump(schedule_config, config_file)


def schedule(
    *,
    # ___________________________________ SLURM #
    working_dir: str = '.',
    jobs_dir: str = 'jobs',
    partition: str = 'publicgpu',
    account: str = 'miv',
    node: int = 1,
    task: int = 1,
    cpu: int = 1,
    gpu: int = 1,
    ram: int = 16,
    constraint: str = 'gpua100|gpurtx6000|gpurtx5000|gpuv100',
    modules: list[str] = ['python/Anaconda3-2019', 'cuda/cuda-11.8', 'gcc/gcc-11'],
    commands: list[str] = ['source /usr/local/Anaconda/Anaconda3-2019.07/etc/profile.d/conda.sh',
                           'conda deactivate',
                           'conda activate torch2cu118'],
    # _____________________________________ JOB #
    file: str,
    function: str,
    args: dict[Any: Any] = dict(),
    # ________________________________ SCHEDULE #
    name: str,
    schedule: dict[str: Sequence[Any]] = dict(),
) -> None:
    config_path = parse_config_path_arg()
    if config_path is not None:
        run(config_path)
        return
    params = locals()
    params.pop('config_path')
    create_configs(**params)
    schedule_config_path = Path(working_dir).resolve() / 'schedule.toml'
    HPCScheduler(config_path=schedule_config_path).run()



from __future__ import annotations
from typing import Any
import os
import toml
import shutil
from pathlib import Path
from copy import deepcopy


class HPCScheduler:

    def __init__(self, config_path: str) -> None:
        self.config_path = config_path
        self.config = toml.load(config_path)
        self.template = toml.load(self.template_path)

    @property
    def parent(self) -> Path:
        return Path(self.config_path).resolve().parent

    @property
    def jobs_dir(self) -> Path:
        return Path(self.parent / self.config['slurm']['jobs_dir']).resolve()

    @property
    def template_path(self) -> Path:
        return self.parent / 'template.toml'

    @property
    def slurm_script_params(self) -> tuple[str | int | list[str]]:
        return (self.config['slurm']['working_dir'],
                self.config['slurm']['jobs_dir'],
                self.config['slurm']['partition'],
                self.config['slurm']['account'],
                self.config['slurm']['node'],
                self.config['slurm']['task'],
                self.config['slurm']['cpu'],
                self.config['slurm']['gpu'],
                self.config['slurm']['ram'],
                self.config['slurm']['constraint'],
                self.config['slurm']['modules'],
                self.config['slurm']['commands'])

    def setup_jobs_env(self) -> None:
        self.jobs_dir.mkdir(exist_ok=True)
        shutil.copyfile(self.template_path, self.jobs_dir / 'template.toml')
        shutil.copyfile(self.config_path,   self.jobs_dir / 'schedule.toml')

    def write_config(self, key: str, value: Any) -> Path:
        config = deepcopy(self.template)
        name = f"{self.config['slurm']['name']}_{key}={value}"
        config['name'] = name
        config['args'][key] = value
        path = (self.jobs_dir / name).with_suffix('.toml')
        with open(path, 'w') as file:
            toml.dump(config, file)
        return path

    def write_configs(self) -> list[Path]:
        config_paths = list()
        for key, values in self.config['parameters'].items():
            for value in values:
                path = self.write_config(key, value)
                config_paths.append(path)
        return config_paths

    def prepare_jobs(self) -> list[Path]:
        self.setup_jobs_env()
        return self.write_configs()

    def write_slurm_job_files(self, config_paths: list[Path]) -> None:
        for config_path in config_paths:
            config = toml.load(config_path)
            job_name = config['name']
            script = get_slurm_script(job_name, *self.slurm_script_params)
            script_path = str(self.jobs_dir / f'{job_name}.job')
            with open(script_path, 'w') as file:
                file.write(script)

    def submit_slurm_schedule(self) -> None:
        jobs_paths = list(self.jobs_dir.glob('*.job'))
        for path in jobs_paths:
            os.system(f'chmod +x {path}')
            os.system(f'sbatch {path}')

    def run(self) -> None:
        config_paths = self.prepare_jobs()
        self.write_slurm_job_files(config_paths)
        self.submit_slurm_schedule()


# _______________________________________________________________________________________________ #


def get_slurm_script(
    job_name: str,
    working_dir: str,
    jobs_dir: str,
    partition: str,
    account: str,
    node: int,
    task: int,
    cpu: int,
    gpu: int,
    ram: int,
    constraint: str,
    modules: list[str],
    commands: list[str],
) -> str:
    working_dir = str(Path(working_dir).resolve())
    stem = Path(jobs_dir) / job_name
    config, output = f'{stem}.toml', f'{stem}.out'
    slurm = f"""#! /bin/bash


# +------------------------------------------------------------------------------------+ #
# |                                  SLURM PARAMETERS                                  | #
# +------------------------------------------------------------------------------------+ #

#SBATCH -p {partition} -A {account}
#SBATCH -N {node}
#SBATCH -n {task}
#SBATCH -c {cpu}
#SBATCH --gres=gpu:{gpu}
#SBATCH --mem={ram}G
#SBATCH --constraint="{constraint}"
#SBATCH -o {output}

"""
    env = """
# +------------------------------------------------------------------------------------+ #
# |                                ENVIRONNEMENT SET UP                                | #
# +------------------------------------------------------------------------------------+ #

"""
    for module in modules:
        env += f"module load {module}\n"
    for command in commands:
        env += f"{command}\n"
    env += f"cd {working_dir}\n\n"
    run = f"""
# +------------------------------------------------------------------------------------+ #
# |                                 RUN PYTHON SCRIPT                                  | #
# +------------------------------------------------------------------------------------+ #

python schedule.py --config {config}
"""
    return slurm + env + run



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup
from tessti import __author__, __email__, __version__


name = "tessti"
url = f"https://github.com/ListIndexOutOfRange/{name}"

with open("README.md", "r") as f:
    readme = f.read()

install_requires = ["toml",]

setup(
    name=name,
    version=__version__,
    description="A tool to schedule SLURM jobs made as simple as possible.",
    long_description_content_type="text/markdown",
    long_description=readme,
    author=__author__,
    author_email=__email__,
    url=url,
    packages=["tessti", ],
    package_dir={"tessti": "tessti"},
    install_requires=install_requires,
    license="GPLv3",
    zip_safe=False,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
    ],
    python_requires=">=3.7",
)


# _______________________________________________________________________________________________ #
# Using this script:
# >>> python setup.py sdist bdist_wheel

# Resulting package should be inspected, checked, and tested with:
# > twine check dist/*
# > twine upload --repository-url https://test.pypi.org/legacy/ dist/*

# Then finally if everything is ok:
# twine upload dist/*
# _______________________________________________________________________________________________ #
